import numpy as np
from OpenGL.GL import *
from pyrr import Vector3

import external


class HealPix(object):

    sampler = None

    @classmethod
    def check_array(cls, array):
        if array.dtype != np.float32:
            return 1
        if array.ndim != 4:
            return 2
        (face_count, height, width, depth) = array.shape
        if depth != 4:
            return 3
        if face_count != 12:
            return 4
        if width != height:
            return 5
        return 0

    @classmethod
    def check_image(cls, image):
        if image.dtype != np.float32:
            return 1
        if image.ndim != 3:
            return 2
        (height, width, depth) = image.shape
        if depth != 4:
            return 3
        if 3*width != 4*height:
            return 4
        return 0

    @classmethod
    def from_array(cls, faces):
        assert cls.check_array(faces) == 0
        return HealPix(faces)

    @classmethod
    def from_image(cls, image):
        assert cls.check_image(image) == 0
        # decompose 4x3 mosaic into healpix faces
        (pos, mid, neg) = [np.hsplit(ring, 4) for ring in np.vsplit(image, 3)]
        array = np.stack(pos + mid + neg)
        return HealPix(array)

    def to_image(self):
        # combine cube map faces into a 3x2 mosaic
        return np.vstack([
            np.hstack([self.array[0], self.array[1], self.array[2], self.array[3]]),
            np.hstack([self.array[4], self.array[5], self.array[6], self.array[7]]),
            np.hstack([self.array[8], self.array[9], self.array[10], self.array[11]]),
        ])

    def to_gl(self, texture_id):
        # enable cube map texturing
        glEnable(GL_TEXTURE_2D)
        # generate and bind texture
        glBindTexture(GL_TEXTURE_2D_ARRAY, texture_id)
        # set up texturing parameters
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_GENERATE_MIPMAP, GL_FALSE)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_BASE_LEVEL, 0)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAX_LEVEL, 0)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)
        (faces, height, width, depth) = self.array.shape
        glTexImage3D(
            GL_TEXTURE_2D_ARRAY, 0, 4,
            width, height, 12, 0,
            GL_RGBA, GL_FLOAT, self.array,
        )

    def get_glsl_sampler(self):
        return '''
uniform sampler2DArray healpixMap;
const float M_PI = 3.141592653589793238462643383;
vec4 sample(vec3 v) {
    float az = 3 * abs(v.z);
    float phi = 2 * atan(v.y, v.x) / M_PI + 2;
    float t, u, face;
    if (az < 2) {
        float x = phi + 0.5;
        float y = v.z * 0.75;
        float t_i, u_i;
        t = modf(x + y, t_i);
        u = modf(x - y, u_i);
        face = min(t_i, u_i);
        face += 4 * (1 + sign(u_i - t_i) - step(4, face));
    } else {
        float phi_i;
        float phi_f = modf(phi, phi_i);
        float s = sqrt(3 - az);
        float dir = step(v.z, 0);
        t = mix(1 - dir, phi_f, s);
        u = mix(dir, phi_f, s);
        face = phi_i + 8 * dir;
    }
    return texture(healpixMap, vec3(t, u, face));
}
'''

    def bind_glsl_texture(self, texture_id, shader):
        glBindTexture(GL_TEXTURE_2D_ARRAY, texture_id)
        shader.uniformi('healpixMap', texture_id)

    @classmethod
    def from_sphere(cls, sphere, resolution=None):
        pass

    def __init__(self, faces):
        # sanity check
        #assert external.healpix_check(faces) == 0
        self.array = faces
        self.resolution = int(12 * faces.shape[1]**2)



import argparse
from ctypes import *
import numpy as np
from OpenGL.GL import *
from OpenGL.GLU import gluPerspective
from PIL import Image
from pyrr import Quaternion, Vector3, Matrix44
from PySide import QtCore
from PySide.QtGui import QMainWindow, QApplication, QDockWidget
from PySide.QtOpenGL import QGLWidget

import projection as proj
from geom import SphericalMesh
from glsl import Shader
from layer import Layer, LayerList, LayerListWithToolBar

class SphairaView(QGLWidget):

    def __init__(self, layers):
        super(SphairaView, self).__init__()
        self.mesh = SphericalMesh(4)
        self.old_pos = QtCore.QPoint(0, 0)
        self.setMouseTracking(True)
        self.layers = layers
        self.zoom = 0.9

    def load_file(self, filename, in_format):
        layer = Layer()
        layer.load_file(filename, in_format)
        self.layers.add_layer(layer)

    def initializeGL(self):
        pass

    def resizeGL(self, w, h):
        glViewport(0, 0, w, h)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPerspective(50, float(w) / h, .01, 100)

    def mouseMoveEvent(self, event):
        pos = event.pos()
        # rotate on left-drag
        if event.buttons() & QtCore.Qt.LeftButton > 0:
            # the rotation vector is the displacement vector rotated by 90 degrees
            dx = pos.x() - self.old_pos.x()
            dy = pos.y() - self.old_pos.y()
            if dx == 0 and dy == 0:
                return
            v = Vector3([dy, dx, 0])
            # update the current orientation
            self.layers.multiplyOrientation(Quaternion.from_axis_rotation(
                -v.normalised,
                -v.length * 0.002,
            ))
        elif event.buttons() & QtCore.Qt.RightButton > 0:
            dz = pos.y() - self.old_pos.y()
            self.zoom = max(0, self.zoom + dz / 100.0)
        self.old_pos = pos
        self.update()

    def paintGL(self):
        glClearColor(0, 0, 0, 1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glEnable(GL_CULL_FACE)
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        glFrontFace(GL_CCW)
        glCullFace(GL_BACK if self.zoom > 1.0 else GL_FRONT)
        for layer in self.layers:
            glMatrixMode(GL_MODELVIEW)
            glLoadIdentity()
            glTranslatef(0, 0, -self.zoom)
            # draw stuff
            layer.shader.bind()
            layer.shader.uniformf('alphaFactor', layer.alpha())
            m = layer.orientation.matrix33
            array = (GLdouble * 9)()
            for i in xrange(3):
                for j in xrange(3):
                    array[3*i + j] = m[i,j]
            layer.shader.uniformf_m3x3('orientation', array)
            glActiveTexture(GL_TEXTURE0 + layer.texture_id)
            layer.sphere.bind_glsl_texture(layer.texture_id, layer.shader)
            self.mesh.draw_triangles()


class SphairaApp(QApplication):

    def __init__(self, args):
        super(SphairaApp, self).__init__(args)
        name = 'Sphaira Viewer'
        self.setApplicationName(name)
        self.layer_widget = QDockWidget('Layers')
        self.layer_widget.setFeatures(
            QDockWidget.DockWidgetMovable |
            QDockWidget.DockWidgetFloatable
        )
        self.layer_list = LayerListWithToolBar()
        self.layer_widget.setWidget(self.layer_list)
        self.gl_widget = SphairaView(self.layer_list)
        self.mainWindow = QMainWindow()
        self.mainWindow.setWindowTitle(name)
        self.mainWindow.setCentralWidget(self.gl_widget)
        self.mainWindow.addDockWidget(
            QtCore.Qt.DockWidgetArea.BottomDockWidgetArea,
            self.layer_widget,
        )
        self.mainWindow.resize(800, 600)
        self.mainWindow.showMaximized()

    def load_file(self, file_name, in_format):
        return self.gl_widget.load_file(file_name, in_format)


def main():
    parser = argparse.ArgumentParser(
        prog='view',
        description='Sphaira viewer for spherical data.',
    )
    parser.add_argument('-i', '--in_format', help='IN_FORMAT')
    parser.add_argument('input', nargs='*', help='INPUT')
    (args, leftover) = parser.parse_known_args()
    in_format = proj.get_format(args.in_format)
    app = SphairaApp(leftover)
    for filename in args.input:
        app.load_file(filename, in_format)
    app.exec_()


if __name__ == '__main__':
    main()



# Based on code by Tristam Macdonald.

from OpenGL.GL import *
from OpenGL.GL import shaders
from ctypes import *

class Shader:

    def __init__(self, vert = [], frag = []):
        self.handle = shaders.compileProgram(
            shaders.compileShader(vert, GL_VERTEX_SHADER),
            shaders.compileShader(frag, GL_FRAGMENT_SHADER),
        )

    def bind(self):
        # bind the program
        glUseProgram(self.handle)

    def unbind(self):
        # unbind whatever program is currently bound - not necessarily this program,
        # so this should probably be a class method instead
        glUseProgram(0)

    # upload a floating point uniform
    # this program must be currently bound
    def uniformf(self, name, *vals):
        # check there are 1-4 values
        if len(vals) in range(1, 5):
            # select the correct function
            { 1 : glUniform1f,
                2 : glUniform2f,
                3 : glUniform3f,
                4 : glUniform4f
                # retrieve the uniform location, and set
            }[len(vals)](glGetUniformLocation(self.handle, name), *vals)

    # upload an integer uniform
    # this program must be currently bound
    def uniformi(self, name, *vals):
        # check there are 1-4 values
        if len(vals) in range(1, 5):
            # select the correct function
            { 1 : glUniform1i,
                2 : glUniform2i,
                3 : glUniform3i,
                4 : glUniform4i
                # retrieve the uniform location, and set
            }[len(vals)](glGetUniformLocation(self.handle, name), *vals)

    # upload a 3x3 matrix uniform
    def uniformf_m3x3(self, name, mat):
        # obtian the uniform location
        loc = glGetUniformLocation(self.handle, name)
        # uplaod the 3x3 floating point matrix
        glUniformMatrix3fv(loc, 1, False, (c_float * 9)(*mat))
